from .aaindex1 import *
# from .aaindex2 import *
# from .aaindex3 import *

#aaindex software metadata
__name__ = "aaindex"
__version__ = "1.0.5"
__description__ = "aaindex is a lightweight Python software package for accessing the data in the various AAIndex databases, which represent the physiochemical and biochemical properties of amino acids as numerical indices."
__author__ = 'AJ McKenna, https://github.com/amckenna41'
__authorEmail__ = 'amckenna41@qub.ac.uk'
__license__ = 'MIT'
__url__ = 'https://github.com/amckenna41/aaindex'
__download_url__ = "https://github.com/amckenna41/aaindex/archive/refs/heads/main.zip"
__status__ = "Production"
__maintainer__ = "AJ McKenna"
__keywords__ = ["amino acid index", "bioinformatics", "protein engineering", "python", "pypi", \
        "physiochemical properties", "biochemical properties", "proteins", "psp", "pysar"]
__test_suite__ = "tests"